from os import system
from os.path import dirname, realpath, basename, join
from glob import glob
import importlib
import logging

from rocolib import rocopath
from rocolib.utils.io import load_yaml
from rocolib.api.components import Component


log = logging.getLogger(__name__)
ROCOLIB_LIBRARY = dirname(realpath(__file__))

pyComponents = [ basename(f)[:-3] for f in glob(ROCOLIB_LIBRARY + "/[!_]*.py")]
yamlComponents = [ basename(f)[:-5] for f in glob(ROCOLIB_LIBRARY + "/*.yaml")]
allComponents = list(set(pyComponents + yamlComponents))

def getComponent(c, **kwargs):
  '''
  Here we are  doing Dynamic instantiation from string name of a class in dynamically imported module
  Parameter c (str): component name e.g. 'Stool'
  '''
  if c in pyComponents:
    # Load "module.submodule.MyClass"
    obj = getattr(importlib.import_module("rocolib.library." + c), c)
    # Instantiate the class (pass arguments to the constructor, if needed)
    my_obj = obj()
  elif c in yamlComponents:
    my_obj = Component(f"{ROCOLIB_LIBRARY}/{c}.yaml")
  else:
    raise ValueError(f"Component {c} not found in library")

  for k, v in kwargs.items():
    if k == 'name':
      my_obj.setName(v)
    else:
      my_obj.setParameter(k, v)
  if 'name' not in kwargs:
    my_obj.setName(c)

  return my_obj

def rebuild(built=None):
    if built is None:
        built = set()
    success = True
    for c in yamlComponents:
        success &= rebuildComponent(c, built, throw=False)
    assert success, "Error rebuilding all components, check stdout/stderr for details"

def rebuildComponent(c, built=None, throw=True):
    if c not in yamlComponents:
        log.debug(f"{c} is not a yaml Component, skipping")
        return True
    if built is None:
        built = set()
    if c in built:
        log.debug(f"{c} has been rebuilt, skipping")
        return True

    definition = load_yaml(c)
    src = definition.get("source", None)
    success = True
    if src:
        log.info(f"Rebuilding {c} from {ROCOLIB_LIBRARY}/{src}...")

        subcomponents = set()
        for k, v in definition.get("subcomponents", dict()).items():
            subcomponents.add(v["classname"])
        for sc in subcomponents:
            rebuildComponent(sc, built)

        # XXX TOOD: Test to make sure we don't call this script and then infinitely recurse!
        log.debug(f"Calling os.system: % python {ROCOLIB_LIBRARY}/{src}")
        if system(f"python {ROCOLIB_LIBRARY}/{src}"):
            success = False

        built.add(c)
        log.debug(repr(built))
        log.debug(f"Done rebuilding {c}.")
    if throw:
        assert success, f"Error rebuilding {c}, check stdout/stderr for details"
    return success

def getComponentPaths(c):
    paths = {}
    if c in pyComponents:
        paths["python"] = rocopath(join(ROCOLIB_LIBRARY, f"{c}.py"))
    if c in yamlComponents:
        paths["yaml"] = rocopath(join(ROCOLIB_LIBRARY, f"{c}.yaml"))
        definition = load_yaml(c)
        src = definition.get("source", None)
        if src:
            paths["builder"] = rocopath(join(ROCOLIB_LIBRARY, src))
    return paths


# tag : [[required ports], [forbidden ports]]
tagDefinitions = {
  'sensor': [["DataOutputPort"],[]],
  'actuator': [["DataInputPort"],[]],
  'mechanical': [["EdgePort"],[]],
  'device': [["MountPort"],[]],
  'UI': [[],["MountPort", "EdgePort"]]
}

def tag(ports):
  tags = {}
  portset = set(ports.keys())
  for tag, (must, cant) in tagDefinitions.items():
    if set(must).issubset(portset) and not len(set(cant).intersection(portset)):
      tags[tag] = [port for ptype in must for port in ports[ptype] ]
  return tags

_taggedComponents = {}
def getTags(x):
  if x in _taggedComponents:
    return _taggedComponents[x]

  try:
    c = getComponent(x)
  except:
    return None

  if isinstance(c, Component):
    interfaces = list(c.interfaces.keys())
    ports = {}
    for iname in interfaces:
      i = c.getInterface(iname)
      iclass = i.__class__.__name__
      try:
        ports[iclass].append(iname)
      except KeyError:
        ports[iclass] = [iname]
    _taggedComponents[x] = tag(ports)
    return tag(ports)
  return None

def taggedComponents(components = None):
  if components == None:
    components = allComponents
  for x in components:
    if getTags(x):
      yield x, getTags(x)

def filterComponents(tagList, components = None):
  for x, tags in taggedComponents(components):
    if set(tagList).issubset(set(tags.keys())):
      yield x, [port for tag in tagList for port in tags[tag] ]
